import yaml
import re
from collections import OrderedDict
from . import table


class ColumnBuilder:
    def __init__(self, ctx, result_list):
        self.ctx = ctx
        self.result_list = result_list

    def _get_column_names(self, subtable_column_name):
        results = set()
        if not self.result_list:
            return results
        obj = self.result_list[0]
        if subtable_column_name:
            obj = getattr(obj, subtable_column_name)
            if not obj:
                # nothing available
                return results
            obj = obj[0]
        return set(obj.to_dict().keys())

    def _make_subtable(self, table_name, format_doc):
        m = re.search("(.*)\\((.+)\\).*", table_name)
        all_options = {}
        if m:
            found = m.group(2)
            for option in found.split(","):
                split_str = option.split("=")
                all_options[split_str[0]] = split_str[1]
            table_name = m.group(1)
        columns = self.build_columns(format_doc, subtable_column_name=table_name)
        return table.subtable(
            self.ctx,
            table_name,
            columns=columns,
            optional=True,
            **all_options,
        )

    def _get_output_column(self, column_name: str):
        m = re.search("(.*)\\((.+)\\).*", column_name)
        if m:
            found = m.group(2)
            all_options = {}
            for option in found.split(","):
                split_str = option.split("=")
                all_options[split_str[0]] = split_str[1]
            return table.column(m.group(1), optional=True, **all_options)
        return table.column(column_name, optional=True)

    def build_columns(
        self,
        format_doc,
        subtable_column_name=None,
        **kwargs,
    ):
        column_names = OrderedDict()
        for column_name in format_doc:
            if column_name == "all":
                for c in self._get_column_names(subtable_column_name):
                    column_names[c] = self._get_output_column(c)
            elif isinstance(column_name, dict):
                sub_table_col = list(column_name.keys())[0]
                sub_table_val = column_name[sub_table_col]
                column_names[sub_table_col] = self._make_subtable(
                    sub_table_col,
                    sub_table_val,
                )
            else:
                column_names[column_name] = self._get_output_column(column_name)

        return list(column_names.values())


def make_columns(
    ctx,
    result_list,
    base_columns,
    show=None,
    clear=False,
    **kwargs,
):
    """
    A generic build for making columns, configurable by user
    to choose specific columns or subtables.

    Add the following options to your command:
    @click.option("--show-columns", type=str, default=None)
    @click.option("--reset-columns", is_flag=True, default=False)

    Then call make_columns (for example)
    columns = make_columns(ctx, results, ["id", "email"], show=show, clear=clear)

    This make the 'default' show have columns 'id', 'email'.

    You may clear it, and start from scratch:

    show just email:
      --clear --show email

    show email and the id
      --clear --show id,email

    show all columns
      --clear --show all

    show a subtable:
      --clear --show "[id,email,upstream_user_identities: [all]]"

    show a specific column in the subtable:
      --clear --show "[id,email,upstream_user_identities: [spec.upstream_idp_id]]"

     rename the output column
      --clear --show "[id,email,upstream_user_identities: [spec.upstream_idp_id(newname=upstream_idp_id)]]"  # noqa

    Note that the show format is a valid yaml doc. Its a list of columns, when
    a mapping (dictionary) is found that is assumed to be a subtable, where the
    key name is the subtable name
    """
    format_doc = []
    if not clear and base_columns:
        format_doc = yaml.safe_load(base_columns)

    if show:
        if "[" not in show:
            # incase the user hasn't created a valid yaml list,
            # make it a list.
            show = "[" + show + "]"
        format_doc.extend(yaml.safe_load(show))

    return ColumnBuilder(ctx, result_list).build_columns(format_doc, **kwargs)



from unittest.mock import MagicMock

class BillingApiMock:

    def __init__(self):
        self.mock_add_billing_usage_record = MagicMock()
        self.mock_add_customer_balance_transaction = MagicMock()
        self.mock_add_org_to_billing_account = MagicMock()
        self.mock_add_org_to_billing_subscription = MagicMock()
        self.mock_add_subscription_balance_transaction = MagicMock()
        self.mock_create_billing_account = MagicMock()
        self.mock_create_feature = MagicMock()
        self.mock_create_product = MagicMock()
        self.mock_create_subscription = MagicMock()
        self.mock_delete_billing_account = MagicMock()
        self.mock_delete_feature = MagicMock()
        self.mock_delete_product = MagicMock()
        self.mock_delete_subscription = MagicMock()
        self.mock_get_billing_account = MagicMock()
        self.mock_get_billing_account_orgs = MagicMock()
        self.mock_get_billing_subscription_orgs = MagicMock()
        self.mock_get_customer_balance_transactions = MagicMock()
        self.mock_get_feature = MagicMock()
        self.mock_get_product = MagicMock()
        self.mock_get_subscription = MagicMock()
        self.mock_get_subscription_balance_transactions = MagicMock()
        self.mock_get_usage_records = MagicMock()
        self.mock_list_billing_accounts = MagicMock()
        self.mock_list_checkout_sessions = MagicMock()
        self.mock_list_features = MagicMock()
        self.mock_list_products = MagicMock()
        self.mock_list_subscription_features = MagicMock()
        self.mock_list_subscriptions = MagicMock()
        self.mock_list_subscriptions_with_feature = MagicMock()
        self.mock_new_subscription = MagicMock()
        self.mock_remove_org_from_billing_account = MagicMock()
        self.mock_remove_org_from_billing_subscription = MagicMock()
        self.mock_replace_billing_account = MagicMock()
        self.mock_replace_feature = MagicMock()
        self.mock_replace_product = MagicMock()
        self.mock_replace_subscription = MagicMock()
        self.mock_update_subscription_cancellation = MagicMock()

    def add_billing_usage_record(self, *args, **kwargs):
        """
        This method mocks the original api BillingApi.add_billing_usage_record with MagicMock.
        """
        return self.mock_add_billing_usage_record(self, *args, **kwargs)

    def add_customer_balance_transaction(self, *args, **kwargs):
        """
        This method mocks the original api BillingApi.add_customer_balance_transaction with MagicMock.
        """
        return self.mock_add_customer_balance_transaction(self, *args, **kwargs)

    def add_org_to_billing_account(self, *args, **kwargs):
        """
        This method mocks the original api BillingApi.add_org_to_billing_account with MagicMock.
        """
        return self.mock_add_org_to_billing_account(self, *args, **kwargs)

    def add_org_to_billing_subscription(self, *args, **kwargs):
        """
        This method mocks the original api BillingApi.add_org_to_billing_subscription with MagicMock.
        """
        return self.mock_add_org_to_billing_subscription(self, *args, **kwargs)

    def add_subscription_balance_transaction(self, *args, **kwargs):
        """
        This method mocks the original api BillingApi.add_subscription_balance_transaction with MagicMock.
        """
        return self.mock_add_subscription_balance_transaction(self, *args, **kwargs)

    def create_billing_account(self, *args, **kwargs):
        """
        This method mocks the original api BillingApi.create_billing_account with MagicMock.
        """
        return self.mock_create_billing_account(self, *args, **kwargs)

    def create_feature(self, *args, **kwargs):
        """
        This method mocks the original api BillingApi.create_feature with MagicMock.
        """
        return self.mock_create_feature(self, *args, **kwargs)

    def create_product(self, *args, **kwargs):
        """
        This method mocks the original api BillingApi.create_product with MagicMock.
        """
        return self.mock_create_product(self, *args, **kwargs)

    def create_subscription(self, *args, **kwargs):
        """
        This method mocks the original api BillingApi.create_subscription with MagicMock.
        """
        return self.mock_create_subscription(self, *args, **kwargs)

    def delete_billing_account(self, *args, **kwargs):
        """
        This method mocks the original api BillingApi.delete_billing_account with MagicMock.
        """
        return self.mock_delete_billing_account(self, *args, **kwargs)

    def delete_feature(self, *args, **kwargs):
        """
        This method mocks the original api BillingApi.delete_feature with MagicMock.
        """
        return self.mock_delete_feature(self, *args, **kwargs)

    def delete_product(self, *args, **kwargs):
        """
        This method mocks the original api BillingApi.delete_product with MagicMock.
        """
        return self.mock_delete_product(self, *args, **kwargs)

    def delete_subscription(self, *args, **kwargs):
        """
        This method mocks the original api BillingApi.delete_subscription with MagicMock.
        """
        return self.mock_delete_subscription(self, *args, **kwargs)

    def get_billing_account(self, *args, **kwargs):
        """
        This method mocks the original api BillingApi.get_billing_account with MagicMock.
        """
        return self.mock_get_billing_account(self, *args, **kwargs)

    def get_billing_account_orgs(self, *args, **kwargs):
        """
        This method mocks the original api BillingApi.get_billing_account_orgs with MagicMock.
        """
        return self.mock_get_billing_account_orgs(self, *args, **kwargs)

    def get_billing_subscription_orgs(self, *args, **kwargs):
        """
        This method mocks the original api BillingApi.get_billing_subscription_orgs with MagicMock.
        """
        return self.mock_get_billing_subscription_orgs(self, *args, **kwargs)

    def get_customer_balance_transactions(self, *args, **kwargs):
        """
        This method mocks the original api BillingApi.get_customer_balance_transactions with MagicMock.
        """
        return self.mock_get_customer_balance_transactions(self, *args, **kwargs)

    def get_feature(self, *args, **kwargs):
        """
        This method mocks the original api BillingApi.get_feature with MagicMock.
        """
        return self.mock_get_feature(self, *args, **kwargs)

    def get_product(self, *args, **kwargs):
        """
        This method mocks the original api BillingApi.get_product with MagicMock.
        """
        return self.mock_get_product(self, *args, **kwargs)

    def get_subscription(self, *args, **kwargs):
        """
        This method mocks the original api BillingApi.get_subscription with MagicMock.
        """
        return self.mock_get_subscription(self, *args, **kwargs)

    def get_subscription_balance_transactions(self, *args, **kwargs):
        """
        This method mocks the original api BillingApi.get_subscription_balance_transactions with MagicMock.
        """
        return self.mock_get_subscription_balance_transactions(self, *args, **kwargs)

    def get_usage_records(self, *args, **kwargs):
        """
        This method mocks the original api BillingApi.get_usage_records with MagicMock.
        """
        return self.mock_get_usage_records(self, *args, **kwargs)

    def list_billing_accounts(self, *args, **kwargs):
        """
        This method mocks the original api BillingApi.list_billing_accounts with MagicMock.
        """
        return self.mock_list_billing_accounts(self, *args, **kwargs)

    def list_checkout_sessions(self, *args, **kwargs):
        """
        This method mocks the original api BillingApi.list_checkout_sessions with MagicMock.
        """
        return self.mock_list_checkout_sessions(self, *args, **kwargs)

    def list_features(self, *args, **kwargs):
        """
        This method mocks the original api BillingApi.list_features with MagicMock.
        """
        return self.mock_list_features(self, *args, **kwargs)

    def list_products(self, *args, **kwargs):
        """
        This method mocks the original api BillingApi.list_products with MagicMock.
        """
        return self.mock_list_products(self, *args, **kwargs)

    def list_subscription_features(self, *args, **kwargs):
        """
        This method mocks the original api BillingApi.list_subscription_features with MagicMock.
        """
        return self.mock_list_subscription_features(self, *args, **kwargs)

    def list_subscriptions(self, *args, **kwargs):
        """
        This method mocks the original api BillingApi.list_subscriptions with MagicMock.
        """
        return self.mock_list_subscriptions(self, *args, **kwargs)

    def list_subscriptions_with_feature(self, *args, **kwargs):
        """
        This method mocks the original api BillingApi.list_subscriptions_with_feature with MagicMock.
        """
        return self.mock_list_subscriptions_with_feature(self, *args, **kwargs)

    def new_subscription(self, *args, **kwargs):
        """
        This method mocks the original api BillingApi.new_subscription with MagicMock.
        """
        return self.mock_new_subscription(self, *args, **kwargs)

    def remove_org_from_billing_account(self, *args, **kwargs):
        """
        This method mocks the original api BillingApi.remove_org_from_billing_account with MagicMock.
        """
        return self.mock_remove_org_from_billing_account(self, *args, **kwargs)

    def remove_org_from_billing_subscription(self, *args, **kwargs):
        """
        This method mocks the original api BillingApi.remove_org_from_billing_subscription with MagicMock.
        """
        return self.mock_remove_org_from_billing_subscription(self, *args, **kwargs)

    def replace_billing_account(self, *args, **kwargs):
        """
        This method mocks the original api BillingApi.replace_billing_account with MagicMock.
        """
        return self.mock_replace_billing_account(self, *args, **kwargs)

    def replace_feature(self, *args, **kwargs):
        """
        This method mocks the original api BillingApi.replace_feature with MagicMock.
        """
        return self.mock_replace_feature(self, *args, **kwargs)

    def replace_product(self, *args, **kwargs):
        """
        This method mocks the original api BillingApi.replace_product with MagicMock.
        """
        return self.mock_replace_product(self, *args, **kwargs)

    def replace_subscription(self, *args, **kwargs):
        """
        This method mocks the original api BillingApi.replace_subscription with MagicMock.
        """
        return self.mock_replace_subscription(self, *args, **kwargs)

    def update_subscription_cancellation(self, *args, **kwargs):
        """
        This method mocks the original api BillingApi.update_subscription_cancellation with MagicMock.
        """
        return self.mock_update_subscription_cancellation(self, *args, **kwargs)




from . import context
from .input_helpers import get_org_from_input_or_ctx
import agilicus


def lookup(ctx, guid, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    org_id = get_org_from_input_or_ctx(ctx, **kwargs)

    res = apiclient.lookups_api.lookup_org_guid(org_id, guid)
    return res


def bulk_lookup(ctx, guids, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    org_id = get_org_from_input_or_ctx(ctx, **kwargs)

    request = agilicus.LookupRequest(org_id=org_id, guids=guids)

    res = apiclient.lookups_api.bulk_query_org_guids(request)
    return res



import operator
import json
from babel import numbers
from datetime import timezone, datetime

from agilicus import context
from agilicus.output.json import convert_to_json

from dataclasses import dataclass
from typing import Callable

from prettytable import PrettyTable, TableStyle
from . import column_builder


def format_date(date_input):
    return date_input.strftime("%Y-%m-%d %H:%M:%S.%f")


def format_date_only(date_input):
    return date_input.strftime("%Y-%m-%d")


def date_else_identity(input_obj, ctx=None):
    if isinstance(input_obj, datetime):
        return format_date(input_obj)
    if ctx and not context.output_console(ctx):
        return input_obj
    if isinstance(input_obj, list):
        list_as_str = ""
        for obj in input_obj:
            try_dict = None
            try:
                try_dict = obj.to_dict()
            except Exception:
                pass
            if isinstance(try_dict, dict):
                prefix = "-"
                for k, v in try_dict.items():
                    list_as_str += f"{prefix} {k}: {v}\n"
                    prefix = " "
            else:
                list_as_str += f"- {obj}\n"
        return list_as_str
    return input_obj


def format_timestamp(input_obj, ctx=None):
    return datetime.fromtimestamp(input_obj)


def format_currency_from_cents(cents, currency=None, ctx=None):
    currency_symbol = ""
    if currency:
        currency_symbol = numbers.get_currency_symbol(currency.upper(), locale="en")
    if cents >= 0:
        return currency_symbol + "{:,.2f}".format(cents / 100)
    else:
        return "(" + currency_symbol + "{:,.2f}".format(-cents / 100) + ")"


@dataclass
class OutputColumn:
    in_name: str
    out_name: str
    format_fn: Callable = date_else_identity
    getter: Callable = None
    optional: bool = False


def column(name, newname=None, optional=False, getter=None, format_fn=None):
    out_name = newname
    if not out_name:
        out_name = name
    if not format_fn:
        format_fn = date_else_identity
    return OutputColumn(
        in_name=name,
        out_name=out_name,
        optional=optional,
        getter=getter,
        format_fn=format_fn,
    )


def mapped_column(in_name, out_name):
    return OutputColumn(in_name=in_name, out_name=out_name)


def subtable(
    ctx,
    in_name,
    columns,
    out_name=None,
    subobject_name=None,
    table_getter=operator.attrgetter,
    optional=False,
):
    if not out_name:
        out_name = in_name

    outer_ctx = ctx

    def format_fn(records, ctx=None):
        return format_table(
            outer_ctx, records, columns, getter=table_getter, nested=True
        )

    def getter(record, base_getter):
        subobject = base_getter(subobject_name)(record)
        return base_getter(in_name)(subobject)

    col_getter = None
    if subobject_name:
        col_getter = getter

    return OutputColumn(
        in_name=in_name,
        out_name=out_name,
        format_fn=format_fn,
        getter=col_getter,
        optional=optional,
    )


def subobject_column(
    in_name,
    out_name,
    subobject_name,
    optional=False,
    max_size=None,
    json_dump=None,
    formatter=None,
    subobject_keys=None,
    **kwargs,
):
    if not out_name:
        out_name = in_name

    def output_val(val):
        if json_dump:
            return json.dumps(val, indent=4)
        return val

    def getter(record, base_getter):
        names = subobject_keys
        if not names:
            names = in_name
        if not isinstance(names, list):
            names = [names]

        caught_exc = None
        for name in names:
            val = None
            try:
                subobject = base_getter(subobject_name)(record)
                val = base_getter(name)(subobject)
                if max_size is not None and val:
                    return output_val(val[:max_size])
                if val is not None or len(names) == 1:
                    return output_val(val)
            except Exception as exc:
                caught_exc = exc

        if not optional and caught_exc:
            raise caught_exc
        return None

    result = OutputColumn(
        in_name=in_name, out_name=out_name, getter=getter, optional=optional
    )

    if formatter:
        result.format_fn = formatter
    return result


def metadata_column(in_name, out_name=None, **kwargs):
    return subobject_column(in_name, out_name, "metadata", **kwargs)


def spec_column(in_name, out_name=None, **kwargs):
    return subobject_column(in_name, out_name, "spec", **kwargs)


def status_column(in_name, out_name=None, **kwargs):
    return subobject_column(in_name, out_name, "status", **kwargs)


def constant_if_exists(column: OutputColumn, constant, default=None):
    orig = column.format_fn

    def check(val, ctx=None):
        if orig:
            val = orig(val, ctx=ctx)

        if val:
            return constant
        return default

    column.format_fn = check
    return column


def list_count(column: OutputColumn):
    orig = column.format_fn

    def formatter(val, ctx=None):
        if isinstance(val, list):
            return len(val)
        if orig:
            val = orig(val, ctx=ctx)
        return val

    column.format_fn = formatter
    return column


def list_map(column: OutputColumn, mapfn: Callable):
    orig = column.getter

    def getter(record, getter):
        if orig:
            val = orig(record, getter)
        else:
            val = getter(column.in_name)(record)
        if isinstance(val, list):
            val = [mapfn(v) for v in val]
        return val

    column.getter = getter
    return column


def expand_dict(column: OutputColumn):
    orig = column.getter

    def getter(record, getter):
        if orig:
            val = orig(record, getter)
        else:
            val = getter(column.in_name)(record)
        if val is None:
            return val
        if not isinstance(val, dict):
            val = val.to_dict()
        return list(val.items())

    column.getter = getter
    return column


def summarize(column: OutputColumn, max_length: int):
    orig = column.format_fn

    def format_list(val, ctx=None):
        if len(val) <= max_length:
            if orig:
                val = orig(val, ctx=ctx)
            return val

        val = val[:max_length]
        if orig:
            val = orig(val)
        return val + "[truncated]"

    def formatter(val, ctx=None):
        if isinstance(val, list):
            return format_list(val, ctx=ctx)

        if orig:
            val = orig(val)

        if isinstance(val, str) and len(val) > max_length:
            idx = int(max_length / 2)
            return val[:idx] + "..." + val[-idx:]
        return val

    column.format_fn = formatter
    return column


# An implementation of attrgetter that handles a None in the middle
def short_circuit_attrgetter(item):
    def func(obj):
        for name in item.split("."):
            if obj is None:
                return None
            obj = getattr(obj, name)
        return obj

    return func


def _dump_json(ctx, to_dump, indent=2):
    # Sometimes a compound table can be a dict at its root. When formatting as json
    # handle that case
    if not isinstance(to_dump, list):
        to_dump_as_dict = to_dump
        if not isinstance(to_dump, dict):
            to_dump_as_dict = to_dump.to_dict()
    else:
        to_dump_as_dict = [
            record.to_dict() if not isinstance(record, dict) else record
            for record in to_dump
        ]

    return convert_to_json(ctx, to_dump_as_dict, indent=indent)


def output_is_formatted(ctx):
    return context.get_value(ctx, "output_format") in ("console", "csv")


def get_table_style(ctx):
    table_style_str = context.get_value(ctx, "output_format")
    table_style = None
    if table_style_str and not output_is_formatted(ctx):
        from prettytable import (
            DEFAULT,
            PLAIN_COLUMNS,
            TableStyle,
        )

        if table_style_str == "default":
            table_style = DEFAULT
        elif table_style_str == "markdown":
            table_style = TableStyle.MARKDOWN
        elif table_style_str == "plain":
            table_style = PLAIN_COLUMNS
        elif table_style_str == "msword":
            table_style = TableStyle.MSWORD_FRIENDLY
        else:
            raise Exception(f"invalid table style {table_style_str}")
    return table_style


def format_table(
    ctx,
    records,
    columns,
    getter=short_circuit_attrgetter,
    row_filter=None,
    nested=False,
):
    if context.output_json(ctx):
        return _dump_json(ctx, records)
    table = PrettyTable([column.out_name for column in columns])

    table_style = get_table_style(ctx)
    if table_style:
        table.set_style(table_style)

    if not records:
        records = []

    if not isinstance(records, list):
        records = [records]

    for record in records:
        row = []

        if row_filter is not None:
            if not row_filter(record):
                continue
        for column in columns:
            in_value = None
            if column.getter:
                try:
                    in_value = column.getter(record, getter)
                except Exception as exc:
                    if not column.optional:
                        raise exc
            else:
                try:
                    in_value = getter(column.in_name)(record)
                except Exception as exc:
                    if not column.optional:
                        raise exc

            out_value = "---"
            if in_value is not None:
                out_value = column.format_fn(in_value, ctx=ctx)
            row.append(out_value)

        table.add_row(row)
    table.align = "l"

    return _table_to_console(ctx, table, nested)


def _table_to_console(ctx, table: PrettyTable, nested):
    if context.output_csv(ctx):
        if nested:
            return table.get_json_string(header=False, indent=None, default=str)
        return table.get_csv_string()
    if context.output_markdown(ctx):
        table.set_style(TableStyle.MARKDOWN)
    return table


def output_entry(ctx, entry, headers=None):
    if ctx.obj["output_format"] == "json":
        print(_dump_json(ctx, entry))
        if headers and ctx.obj["output_headers"]:
            print(_dump_json(ctx, headers.__dict__))
        return
    table = PrettyTable(["field", "value"])
    if entry:
        for k, v in list(entry.items()):
            if k == "nbf" or k == "exp" or k == "iat":
                _t = v
                if not isinstance(_t, str):
                    _t = datetime.fromtimestamp(v, timezone.utc).strftime(
                        "%Y-%m-%d %H:%M:%S %z"
                    )
                table.add_row([k, json.dumps(_t, indent=4)])
            elif k == "created" or k == "updated":
                table.add_row([k, v])
            elif isinstance(v, str):
                table.add_row([k, v])
            else:
                table.add_row([k, json.dumps(v, indent=4, default=str)])
        table.align = "l"
        print(table)
        if headers and ctx.obj["output_headers"]:
            table = PrettyTable(["header-key", "value"])
            for k, v in list(headers.items()):
                table.add_row([k, v])
            table.align = "l"
            print(table)


def make_columns(*args, **kwargs):
    return column_builder.make_columns(*args, **kwargs)


def _get_with_default(column, record):
    getter = column.getter
    if not getter:
        getter = short_circuit_attrgetter
    return getter(column.in_name)(record)


def object_subtable(
    ctx,
    in_name,
    columns,
    out_name=None,
    subobject_name=None,
    table_getter=operator.itemgetter,
    optional=False,
):
    """
    converts an object with columns defined by "columns" into a subtable.
    """
    if not out_name:
        out_name = in_name

    def scalar_name(column, idx):
        return column.out_name

    def list_name(column, idx):
        return f"{column.out_name}.{idx}"

    subtable_columns = [column("name"), column("value")]

    outer_ctx = ctx

    def format_fn(records, ctx=None):
        subtable_records = []
        namer = scalar_name
        if isinstance(records, list):
            namer = list_name
        else:
            records = [records]

        for idx, record in enumerate(records):
            subtable_records.extend(
                [
                    {
                        "name": namer(column, idx),
                        "value": _get_with_default(column, record),
                    }
                    for column in columns
                ]
            )

        return format_table(
            outer_ctx,
            subtable_records,
            subtable_columns,
            getter=table_getter,
            nested=True,
        )

    def getter(record, base_getter):
        subobject = base_getter(subobject_name)(record)
        return base_getter(in_name)(subobject)

    col_getter = None
    if subobject_name:
        col_getter = getter

    return OutputColumn(
        in_name=in_name,
        out_name=out_name,
        format_fn=format_fn,
        getter=col_getter,
        optional=optional,
    )



"""
    Agilicus API

    Agilicus is API-first. Modern software is controlled by other software, is open, is available for you to use the way you want, securely, simply.  The OpenAPI Specification in YAML format is available on [www](https://www.agilicus.com/www/api/agilicus-openapi.yaml) for importing to other tools.  A rendered, online viewable and usable version of this specification is available at [api](https://www.agilicus.com/api). You may try the API inline directly in the web page. To do so, first obtain an Authentication Token (the simplest way is to install the Python SDK, and then run `agilicus-cli --issuer https://MYISSUER get-token`). You will need an org-id for most calls (and can obtain from `agilicus-cli --issuer https://MYISSUER list-orgs`). The `MYISSUER` will typically be `auth.MYDOMAIN`, and you will see it as you sign-in to the administrative UI.  This API releases on Bearer-Token authentication. To obtain a valid bearer token you will need to Authenticate to an Issuer with OpenID Connect (a superset of OAUTH2).  Your \"issuer\" will look like https://auth.MYDOMAIN. For example, when you signed-up, if you said \"use my own domain name\" and assigned a CNAME of cloud.example.com, then your issuer would be https://auth.cloud.example.com.  If you selected \"use an Agilicus supplied domain name\", your issuer would look like https://auth.myorg.agilicus.cloud.  For test purposes you can use our [Python SDK](https://pypi.org/project/agilicus/) and run `agilicus-cli --issuer https://auth.MYDOMAIN get-token`.  This API may be used in any language runtime that supports OpenAPI 3.0, or, you may use our [Python SDK](https://pypi.org/project/agilicus/), our [Typescript SDK](https://www.npmjs.com/package/@agilicus/angular), or our [Golang SDK](https://git.agilicus.com/pub/sdk-go).  100% of the activities in our system our API-driven, from our web-admin, through our progressive web applications, to all internals: there is nothing that is not accessible.  For more information, see [developer resources](https://www.agilicus.com/developer).   # noqa: E501

    The version of the OpenAPI document: 2025.09.04
    Contact: dev@agilicus.com
    Generated by: https://openapi-generator.tech
"""


import unittest

import agilicus_api
from agilicus_api.api.licensing_api import LicensingApi  # noqa: E501


class TestLicensingApi(unittest.TestCase):
    """LicensingApi unit test stubs"""

    def setUp(self):
        self.api = LicensingApi()  # noqa: E501

    def tearDown(self):
        pass

    def test_create_hypothetical_license_details_query(self):
        """Test case for create_hypothetical_license_details_query

        Query a hypothetical set of license details  # noqa: E501
        """
        pass

    def test_create_license(self):
        """Test case for create_license

        Create a license  # noqa: E501
        """
        pass

    def test_create_product_table_version(self):
        """Test case for create_product_table_version

        Create a product table version  # noqa: E501
        """
        pass

    def test_delete_license(self):
        """Test case for delete_license

        Delete a license  # noqa: E501
        """
        pass

    def test_delete_product_table_version(self):
        """Test case for delete_product_table_version

        Delete a product table version  # noqa: E501
        """
        pass

    def test_get_license(self):
        """Test case for get_license

        Get a single license  # noqa: E501
        """
        pass

    def test_get_product_table_version(self):
        """Test case for get_product_table_version

        Get a single product table version  # noqa: E501
        """
        pass

    def test_list_license_details(self):
        """Test case for list_license_details

        Get all license details  # noqa: E501
        """
        pass

    def test_list_license_evaluation_contexts(self):
        """Test case for list_license_evaluation_contexts

        Get license evaluation context  # noqa: E501
        """
        pass

    def test_list_licenses(self):
        """Test case for list_licenses

        Get all licenses  # noqa: E501
        """
        pass

    def test_list_product_table_versions(self):
        """Test case for list_product_table_versions

        Get all product table versions  # noqa: E501
        """
        pass

    def test_replace_license(self):
        """Test case for replace_license

        Create or update a license  # noqa: E501
        """
        pass

    def test_replace_product_table_version(self):
        """Test case for replace_product_table_version

        Create or update a product table version  # noqa: E501
        """
        pass


if __name__ == '__main__':
    unittest.main()



from operator import attrgetter

from ..output.table import (
    column,
    list_map,
    expand_dict,
    summarize,
)


def constraint_columns(num_subtable_rows=10, column_func=column):
    return [
        summarize(
            list_map(
                column_func("license_constraints", "constraints"), attrgetter("name")
            ),
            num_subtable_rows,
        ),
        summarize(
            list_map(
                expand_dict(column_func("constraint_variables", "variables")),
                lambda item: item[0],
            ),
            num_subtable_rows,
        ),
    ]



from operator import attrgetter
from ..input_helpers import (
    model_from_dict,
    strip_none,
)

from .. import context

from agilicus import (
    create_or_update,
    ProductTableVersion,
)

from ..output.table import (
    format_table,
    spec_column,
    metadata_column,
    constant_if_exists,
    subtable,
    column,
    list_map,
    summarize,
)

from .formatters import constraint_columns


def list_product_table_versions(ctx, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    result = apiclient.licensing_api.list_product_table_versions(
        **strip_none(kwargs),
    )
    return result.product_table_versions


def format_product_table_versions(ctx, templates):
    num_subtable_rows = 10
    feature_columns = [
        column("name"),
        *constraint_columns(num_subtable_rows),
    ]
    product_columns = [
        column("name"),
        *constraint_columns(num_subtable_rows),
        column("included_features", "features"),
    ]
    columns = [
        metadata_column("id"),
        spec_column("version"),
        constant_if_exists(spec_column("published"), "✓", default=""),
        subtable(ctx, "spec.product_table.features", feature_columns, "features"),
        subtable(ctx, "spec.product_table.products", product_columns, "products"),
        summarize(
            list_map(
                spec_column("product_table.global_constraints", "constraints"),
                attrgetter("name"),
            ),
            num_subtable_rows,
        ),
    ]

    return format_table(ctx, templates, columns)


def apply_product_table_version(ctx, input_file, **kwargs):
    version = model_from_dict(ProductTableVersion, input_file)
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    resp, _ = create_or_update(
        version,
        lambda obj: apiclient.licensing_api.create_product_table_version(obj),
        lambda guid, obj: apiclient.licensing_api.replace_product_table_version(
            guid, product_table_version=obj
        ),
        to_dict=False,
    )
    return resp


def delete_product_table_version(ctx, product_table_version_id):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    apiclient.licensing_api.delete_product_table_version(product_table_version_id)


def show_product_table_version(ctx, product_table_version_id, version, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    if not version and not product_table_version_id:
        raise ValueError("version or product_table_version must be provided")
    if version:
        result = apiclient.licensing_api.list_product_table_versions(
            version=version, limit=1
        )
        if not result.product_table_versions:
            raise ValueError(f"version {version} not found")
        return result.product_table_versions[0]

    return apiclient.licensing_api.get_product_table_version(product_table_version_id)



# flake8: noqa
from typing import Callable
import json
import os
import sys
import dataclasses
from . import context
from . import scopes
from . import credentials
from . import tokens
import jwt

from .version import __version__  # noqa

sys.path.append(os.path.dirname(__file__))  # noqa

from .agilicus_api import *  # noqa
from .agilicus_api import exceptions  # noqa
from . import patches  # noqa
from .pagination.pagination import get_many_entries

from .create import create_or_update, add_list_resources, AddInfo, AddResult, find_guid

ApiClient = patches.patched_api_client()


@dataclasses.dataclass
class AgilicusAPIHelper:
    default_org_id: str
    users: UsersApi
    billing: BillingApi
    organisations: OrganisationsApi
    policies: PolicyApi
    certificates: CertificatesApi
    applications: ApplicationsApi
    groups: GroupsApi
    connectors: ConnectorsApi
    resources: ResourcesApi
    resouces: ResourcesApi
    catalogues: CataloguesApi
    permissions: PermissionsApi
    audits: AuditsApi
    files: FilesApi
    tokens: TokensApi
    diagnostics: DiagnosticsApi
    metrics: MetricsApi
    challenges: ChallengesApi
    application_services: ApplicationServicesApi
    issuers: IssuersApi
    messages: MessagesApi
    lookups: LookupsApi
    trusted_certs: TrustedCertsApi
    rules: RulesApi
    policy_config: PolicyConfigApi
    licensing: LicensingApi


def GetClient(
    issuer=context.ISSUER_DEFAULT,
    cacert=None,
    client_id="agilicus-builtin-cli",
    authentication_document=None,
    agilicus_scopes=scopes.DEFAULT_SCOPES,
    auth_local_webserver=True,
    api_url=None,
    expiry=None,
    admin=False,
):

    config = Configuration(host=api_url, ssl_ca_cert=cacert, discard_unknown_keys=True)
    if authentication_document:
        creds = {}
        with open(authentication_document) as fd:
            ad = json.load(fd)
        token = tokens.create_service_token(
            auth_doc=ad,
            scope=agilicus_scopes,
            client_id=client_id,
            expiry=expiry,
            verify=cacert,
        )
        config.access_token = token.get("access_token")
    else:
        creds = credentials.get_credentials(
            issuer=issuer,
            cacert=cacert,
            client_id=client_id,
            agilicus_scopes=agilicus_scopes,
            auth_local_webserver=auth_local_webserver,
            admin=admin,
        )
        config.access_token = creds.access_token

    _default_org_id = None
    access_token = jwt.decode(
        config.access_token,
        algorithms=["ES256"],
        options={"verify_signature": False},
        leeway=60,
    )
    if "org" in access_token:
        _default_org_id = access_token["org"]

    return AgilicusAPIHelper(
        default_org_id=_default_org_id,
        users=UsersApi(ApiClient(config)),
        billing=BillingApi(ApiClient(config)),
        organisations=OrganisationsApi(ApiClient(config)),
        policies=PolicyApi(ApiClient(config)),
        certificates=CertificatesApi(ApiClient(config)),
        applications=ApplicationsApi(ApiClient(config)),
        groups=GroupsApi(ApiClient(config)),
        connectors=ConnectorsApi(ApiClient(config)),
        resouces=ResourcesApi(ApiClient(config)),
        resources=ResourcesApi(ApiClient(config)),
        catalogues=CataloguesApi(ApiClient(config)),
        permissions=PermissionsApi(ApiClient(config)),
        audits=AuditsApi(ApiClient(config)),
        files=FilesApi(ApiClient(config)),
        tokens=TokensApi(ApiClient(config)),
        diagnostics=DiagnosticsApi(ApiClient(config)),
        metrics=MetricsApi(ApiClient(config)),
        challenges=ChallengesApi(ApiClient(config)),
        application_services=ApplicationServicesApi(ApiClient(config)),
        issuers=IssuersApi(ApiClient(config)),
        messages=MessagesApi(ApiClient(config)),
        lookups=LookupsApi(ApiClient(config)),
        trusted_certs=TrustedCertsApi(ApiClient(config)),
        rules=RulesApi(ApiClient(config)),
        policy_config=PolicyConfigApi(ApiClient(config)),
        licensing=LicensingApi(ApiClient(config)),
    )



"""
    Agilicus API

    Agilicus is API-first. Modern software is controlled by other software, is open, is available for you to use the way you want, securely, simply.  The OpenAPI Specification in YAML format is available on [www](https://www.agilicus.com/www/api/agilicus-openapi.yaml) for importing to other tools.  A rendered, online viewable and usable version of this specification is available at [api](https://www.agilicus.com/api). You may try the API inline directly in the web page. To do so, first obtain an Authentication Token (the simplest way is to install the Python SDK, and then run `agilicus-cli --issuer https://MYISSUER get-token`). You will need an org-id for most calls (and can obtain from `agilicus-cli --issuer https://MYISSUER list-orgs`). The `MYISSUER` will typically be `auth.MYDOMAIN`, and you will see it as you sign-in to the administrative UI.  This API releases on Bearer-Token authentication. To obtain a valid bearer token you will need to Authenticate to an Issuer with OpenID Connect (a superset of OAUTH2).  Your \"issuer\" will look like https://auth.MYDOMAIN. For example, when you signed-up, if you said \"use my own domain name\" and assigned a CNAME of cloud.example.com, then your issuer would be https://auth.cloud.example.com.  If you selected \"use an Agilicus supplied domain name\", your issuer would look like https://auth.myorg.agilicus.cloud.  For test purposes you can use our [Python SDK](https://pypi.org/project/agilicus/) and run `agilicus-cli --issuer https://auth.MYDOMAIN get-token`.  This API may be used in any language runtime that supports OpenAPI 3.0, or, you may use our [Python SDK](https://pypi.org/project/agilicus/), our [Typescript SDK](https://www.npmjs.com/package/@agilicus/angular), or our [Golang SDK](https://git.agilicus.com/pub/sdk-go).  100% of the activities in our system our API-driven, from our web-admin, through our progressive web applications, to all internals: there is nothing that is not accessible.  For more information, see [developer resources](https://www.agilicus.com/developer).   # noqa: E501

    The version of the OpenAPI document: 2025.09.04
    Contact: dev@agilicus.com
    Generated by: https://openapi-generator.tech
"""


import re  # noqa: F401
import sys  # noqa: F401

from agilicus_api.api_client import ApiClient, Endpoint as _Endpoint
from agilicus_api.model_utils import (  # noqa: F401
    check_allowed_values,
    check_validations,
    date,
    datetime,
    file_type,
    none_type,
    validate_and_convert_types
)
from agilicus_api.model.error_message import ErrorMessage
from agilicus_api.model.hypothetical_license_details import HypotheticalLicenseDetails
from agilicus_api.model.hypothetical_license_details_query import HypotheticalLicenseDetailsQuery
from agilicus_api.model.license import License
from agilicus_api.model.list_license_details_response import ListLicenseDetailsResponse
from agilicus_api.model.list_license_evaluation_contexts_response import ListLicenseEvaluationContextsResponse
from agilicus_api.model.list_licenses_response import ListLicensesResponse
from agilicus_api.model.list_product_table_versions_response import ListProductTableVersionsResponse
from agilicus_api.model.product_table_version import ProductTableVersion


class LicensingApi(object):
    """NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    def __init__(self, api_client=None):
        if api_client is None:
            api_client = ApiClient()
        self.api_client = api_client

        def __create_hypothetical_license_details_query(
            self,
            hypothetical_license_details_query,
            **kwargs
        ):
            """Query a hypothetical set of license details  # noqa: E501

            Get all per-org license details matching the query. This will return the actual constraints available to an organisation, taking into account their product, the hypothetical version of the product table, and any overrides applied to their organisation. Multiple organisations' constraints may be returned at a time.   # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.create_hypothetical_license_details_query(hypothetical_license_details_query, async_req=True)
            >>> result = thread.get()

            Args:
                hypothetical_license_details_query (HypotheticalLicenseDetailsQuery):

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                HypotheticalLicenseDetails
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['hypothetical_license_details_query'] = \
                hypothetical_license_details_query
            return self.call_with_http_info(**kwargs)

        if self.create_hypothetical_license_details_query is None:
            self.create_hypothetical_license_details_query = _Endpoint(
                settings={
                    'response_type': (HypotheticalLicenseDetails,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/license_details/hypothetical',
                    'operation_id': 'create_hypothetical_license_details_query',
                    'http_method': 'POST',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'hypothetical_license_details_query',
                    ],
                    'required': [
                        'hypothetical_license_details_query',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'hypothetical_license_details_query':
                            (HypotheticalLicenseDetailsQuery,),
                    },
                    'attribute_map': {
                    },
                    'location_map': {
                        'hypothetical_license_details_query': 'body',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [
                        'application/json'
                    ]
                },
                api_client=api_client,
                callable=__create_hypothetical_license_details_query
            )

        def __create_license(
            self,
            license,
            **kwargs
        ):
            """Create a license  # noqa: E501

            Create a license  # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.create_license(license, async_req=True)
            >>> result = thread.get()

            Args:
                license (License):

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                License
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['license'] = \
                license
            return self.call_with_http_info(**kwargs)

        if self.create_license is None:
            self.create_license = _Endpoint(
                settings={
                    'response_type': (License,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/licenses',
                    'operation_id': 'create_license',
                    'http_method': 'POST',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'license',
                    ],
                    'required': [
                        'license',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'license':
                            (License,),
                    },
                    'attribute_map': {
                    },
                    'location_map': {
                        'license': 'body',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [
                        'application/json'
                    ]
                },
                api_client=api_client,
                callable=__create_license
            )

        def __create_product_table_version(
            self,
            product_table_version,
            **kwargs
        ):
            """Create a product table version  # noqa: E501

            Create a product table version  # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.create_product_table_version(product_table_version, async_req=True)
            >>> result = thread.get()

            Args:
                product_table_version (ProductTableVersion):

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                ProductTableVersion
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['product_table_version'] = \
                product_table_version
            return self.call_with_http_info(**kwargs)

        if self.create_product_table_version is None:
            self.create_product_table_version = _Endpoint(
                settings={
                    'response_type': (ProductTableVersion,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/product_table_versions',
                    'operation_id': 'create_product_table_version',
                    'http_method': 'POST',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'product_table_version',
                    ],
                    'required': [
                        'product_table_version',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'product_table_version':
                            (ProductTableVersion,),
                    },
                    'attribute_map': {
                    },
                    'location_map': {
                        'product_table_version': 'body',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [
                        'application/json'
                    ]
                },
                api_client=api_client,
                callable=__create_product_table_version
            )

        def __delete_license(
            self,
            license_id,
            **kwargs
        ):
            """Delete a license  # noqa: E501

            Delete a license  # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.delete_license(license_id, async_req=True)
            >>> result = thread.get()

            Args:
                license_id (str): A license id

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                None
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['license_id'] = \
                license_id
            return self.call_with_http_info(**kwargs)

        if self.delete_license is None:
            self.delete_license = _Endpoint(
                settings={
                    'response_type': None,
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/licenses/{license_id}',
                    'operation_id': 'delete_license',
                    'http_method': 'DELETE',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'license_id',
                    ],
                    'required': [
                        'license_id',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'license_id':
                            (str,),
                    },
                    'attribute_map': {
                        'license_id': 'license_id',
                    },
                    'location_map': {
                        'license_id': 'path',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [],
                },
                api_client=api_client,
                callable=__delete_license
            )

        def __delete_product_table_version(
            self,
            product_table_version_id,
            **kwargs
        ):
            """Delete a product table version  # noqa: E501

            Delete a product table version  # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.delete_product_table_version(product_table_version_id, async_req=True)
            >>> result = thread.get()

            Args:
                product_table_version_id (str): A product table version id

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                None
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['product_table_version_id'] = \
                product_table_version_id
            return self.call_with_http_info(**kwargs)

        if self.delete_product_table_version is None:
            self.delete_product_table_version = _Endpoint(
                settings={
                    'response_type': None,
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/product_table_versions/{product_table_version_id}',
                    'operation_id': 'delete_product_table_version',
                    'http_method': 'DELETE',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'product_table_version_id',
                    ],
                    'required': [
                        'product_table_version_id',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'product_table_version_id':
                            (str,),
                    },
                    'attribute_map': {
                        'product_table_version_id': 'product_table_version_id',
                    },
                    'location_map': {
                        'product_table_version_id': 'path',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [],
                },
                api_client=api_client,
                callable=__delete_product_table_version
            )

        def __get_license(
            self,
            license_id,
            **kwargs
        ):
            """Get a single license  # noqa: E501

            Get a single license  # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.get_license(license_id, async_req=True)
            >>> result = thread.get()

            Args:
                license_id (str): A license id

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                License
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['license_id'] = \
                license_id
            return self.call_with_http_info(**kwargs)

        if self.get_license is None:
            self.get_license = _Endpoint(
                settings={
                    'response_type': (License,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/licenses/{license_id}',
                    'operation_id': 'get_license',
                    'http_method': 'GET',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'license_id',
                    ],
                    'required': [
                        'license_id',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'license_id':
                            (str,),
                    },
                    'attribute_map': {
                        'license_id': 'license_id',
                    },
                    'location_map': {
                        'license_id': 'path',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [],
                },
                api_client=api_client,
                callable=__get_license
            )

        def __get_product_table_version(
            self,
            product_table_version_id,
            **kwargs
        ):
            """Get a single product table version  # noqa: E501

            Get a single product table version  # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.get_product_table_version(product_table_version_id, async_req=True)
            >>> result = thread.get()

            Args:
                product_table_version_id (str): A product table version id

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                ProductTableVersion
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['product_table_version_id'] = \
                product_table_version_id
            return self.call_with_http_info(**kwargs)

        if self.get_product_table_version is None:
            self.get_product_table_version = _Endpoint(
                settings={
                    'response_type': (ProductTableVersion,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/product_table_versions/{product_table_version_id}',
                    'operation_id': 'get_product_table_version',
                    'http_method': 'GET',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'product_table_version_id',
                    ],
                    'required': [
                        'product_table_version_id',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'product_table_version_id':
                            (str,),
                    },
                    'attribute_map': {
                        'product_table_version_id': 'product_table_version_id',
                    },
                    'location_map': {
                        'product_table_version_id': 'path',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [],
                },
                api_client=api_client,
                callable=__get_product_table_version
            )

        def __list_license_details(
            self,
            **kwargs
        ):
            """Get all license details  # noqa: E501

            Get all per-org license details matching the query. This will return the actual constraints available to an organisation, taking into account their product, version of the product table, and any overrides applied to their organisation. Multiple organisations' constraints may be returned at a time, and paged through using page_at_id.   # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.list_license_details(async_req=True)
            >>> result = thread.get()


            Keyword Args:
                limit (int): limit the number of rows in the response. [optional] if omitted the server will use the default value of 500
                org_id (str): Organisation Unique identifier. [optional]
                license_id (str): A license id. [optional]
                license_ids ([str]): A list of license ids. [optional]
                page_at_id (str): Pagination based query with the id as the key. To get the initial entries supply an empty string. On subsequent requests, supply the `page_at_id` field from the list response. . [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                ListLicenseDetailsResponse
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            return self.call_with_http_info(**kwargs)

        if self.list_license_details is None:
            self.list_license_details = _Endpoint(
                settings={
                    'response_type': (ListLicenseDetailsResponse,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/license_details',
                    'operation_id': 'list_license_details',
                    'http_method': 'GET',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'limit',
                        'org_id',
                        'license_id',
                        'license_ids',
                        'page_at_id',
                    ],
                    'required': [],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                        'limit',
                    ]
                },
                root_map={
                    'validations': {
                        ('limit',): {

                            'inclusive_maximum': 500,
                            'inclusive_minimum': 1,
                        },
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'limit':
                            (int,),
                        'org_id':
                            (str,),
                        'license_id':
                            (str,),
                        'license_ids':
                            ([str],),
                        'page_at_id':
                            (str,),
                    },
                    'attribute_map': {
                        'limit': 'limit',
                        'org_id': 'org_id',
                        'license_id': 'license_id',
                        'license_ids': 'license_ids',
                        'page_at_id': 'page_at_id',
                    },
                    'location_map': {
                        'limit': 'query',
                        'org_id': 'query',
                        'license_id': 'query',
                        'license_ids': 'query',
                        'page_at_id': 'query',
                    },
                    'collection_format_map': {
                        'license_ids': 'multi',
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [],
                },
                api_client=api_client,
                callable=__list_license_details
            )

        def __list_license_evaluation_contexts(
            self,
            **kwargs
        ):
            """Get license evaluation context  # noqa: E501

            Get all per-org license evaluation input matching the query. This will return the context dictionary to pass into the evaluation of a license. Multiple organisations' input may be returned at a time, and paged through using page_at_id.   # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.list_license_evaluation_contexts(async_req=True)
            >>> result = thread.get()


            Keyword Args:
                limit (int): limit the number of rows in the response. [optional] if omitted the server will use the default value of 500
                org_id (str): Organisation Unique identifier. [optional]
                license_id (str): A license id. [optional]
                license_ids ([str]): A list of license ids. [optional]
                page_at_id (str): Pagination based query with the id as the key. To get the initial entries supply an empty string. On subsequent requests, supply the `page_at_id` field from the list response. . [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                ListLicenseEvaluationContextsResponse
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            return self.call_with_http_info(**kwargs)

        if self.list_license_evaluation_contexts is None:
            self.list_license_evaluation_contexts = _Endpoint(
                settings={
                    'response_type': (ListLicenseEvaluationContextsResponse,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/licenses/evaluation_contexts',
                    'operation_id': 'list_license_evaluation_contexts',
                    'http_method': 'GET',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'limit',
                        'org_id',
                        'license_id',
                        'license_ids',
                        'page_at_id',
                    ],
                    'required': [],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                        'limit',
                    ]
                },
                root_map={
                    'validations': {
                        ('limit',): {

                            'inclusive_maximum': 500,
                            'inclusive_minimum': 1,
                        },
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'limit':
                            (int,),
                        'org_id':
                            (str,),
                        'license_id':
                            (str,),
                        'license_ids':
                            ([str],),
                        'page_at_id':
                            (str,),
                    },
                    'attribute_map': {
                        'limit': 'limit',
                        'org_id': 'org_id',
                        'license_id': 'license_id',
                        'license_ids': 'license_ids',
                        'page_at_id': 'page_at_id',
                    },
                    'location_map': {
                        'limit': 'query',
                        'org_id': 'query',
                        'license_id': 'query',
                        'license_ids': 'query',
                        'page_at_id': 'query',
                    },
                    'collection_format_map': {
                        'license_ids': 'multi',
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [],
                },
                api_client=api_client,
                callable=__list_license_evaluation_contexts
            )

        def __list_licenses(
            self,
            **kwargs
        ):
            """Get all licenses  # noqa: E501

            Get all Licenses matching the query  # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.list_licenses(async_req=True)
            >>> result = thread.get()


            Keyword Args:
                limit (int): limit the number of rows in the response. [optional] if omitted the server will use the default value of 500
                page_at_id (str): Pagination based query with the id as the key. To get the initial entries supply an empty string. On subsequent requests, supply the `page_at_id` field from the list response. . [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                ListLicensesResponse
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            return self.call_with_http_info(**kwargs)

        if self.list_licenses is None:
            self.list_licenses = _Endpoint(
                settings={
                    'response_type': (ListLicensesResponse,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/licenses',
                    'operation_id': 'list_licenses',
                    'http_method': 'GET',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'limit',
                        'page_at_id',
                    ],
                    'required': [],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                        'limit',
                    ]
                },
                root_map={
                    'validations': {
                        ('limit',): {

                            'inclusive_maximum': 500,
                            'inclusive_minimum': 1,
                        },
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'limit':
                            (int,),
                        'page_at_id':
                            (str,),
                    },
                    'attribute_map': {
                        'limit': 'limit',
                        'page_at_id': 'page_at_id',
                    },
                    'location_map': {
                        'limit': 'query',
                        'page_at_id': 'query',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [],
                },
                api_client=api_client,
                callable=__list_licenses
            )

        def __list_product_table_versions(
            self,
            **kwargs
        ):
            """Get all product table versions  # noqa: E501

            Get all product table versions matching the query  # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.list_product_table_versions(async_req=True)
            >>> result = thread.get()


            Keyword Args:
                limit (int): limit the number of rows in the response. [optional] if omitted the server will use the default value of 500
                version (str): Query based on a version. [optional]
                page_at_version (str): Page based on a version. [optional]
                published (bool): Query based on a whether something is published. [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                ListProductTableVersionsResponse
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            return self.call_with_http_info(**kwargs)

        if self.list_product_table_versions is None:
            self.list_product_table_versions = _Endpoint(
                settings={
                    'response_type': (ListProductTableVersionsResponse,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/product_table_versions',
                    'operation_id': 'list_product_table_versions',
                    'http_method': 'GET',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'limit',
                        'version',
                        'page_at_version',
                        'published',
                    ],
                    'required': [],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                        'limit',
                    ]
                },
                root_map={
                    'validations': {
                        ('limit',): {

                            'inclusive_maximum': 500,
                            'inclusive_minimum': 1,
                        },
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'limit':
                            (int,),
                        'version':
                            (str,),
                        'page_at_version':
                            (str,),
                        'published':
                            (bool,),
                    },
                    'attribute_map': {
                        'limit': 'limit',
                        'version': 'version',
                        'page_at_version': 'page_at_version',
                        'published': 'published',
                    },
                    'location_map': {
                        'limit': 'query',
                        'version': 'query',
                        'page_at_version': 'query',
                        'published': 'query',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [],
                },
                api_client=api_client,
                callable=__list_product_table_versions
            )

        def __replace_license(
            self,
            license_id,
            **kwargs
        ):
            """Create or update a license  # noqa: E501

            Update a license  # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.replace_license(license_id, async_req=True)
            >>> result = thread.get()

            Args:
                license_id (str): A license id

            Keyword Args:
                license (License): [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                License
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['license_id'] = \
                license_id
            return self.call_with_http_info(**kwargs)

        if self.replace_license is None:
            self.replace_license = _Endpoint(
                settings={
                    'response_type': (License,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/licenses/{license_id}',
                    'operation_id': 'replace_license',
                    'http_method': 'PUT',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'license_id',
                        'license',
                    ],
                    'required': [
                        'license_id',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'license_id':
                            (str,),
                        'license':
                            (License,),
                    },
                    'attribute_map': {
                        'license_id': 'license_id',
                    },
                    'location_map': {
                        'license_id': 'path',
                        'license': 'body',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [
                        'application/json'
                    ]
                },
                api_client=api_client,
                callable=__replace_license
            )

        def __replace_product_table_version(
            self,
            product_table_version_id,
            **kwargs
        ):
            """Create or update a product table version  # noqa: E501

            Update a product table version  # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.replace_product_table_version(product_table_version_id, async_req=True)
            >>> result = thread.get()

            Args:
                product_table_version_id (str): A product table version id

            Keyword Args:
                product_table_version (ProductTableVersion): [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                ProductTableVersion
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['product_table_version_id'] = \
                product_table_version_id
            return self.call_with_http_info(**kwargs)

        if self.replace_product_table_version is None:
            self.replace_product_table_version = _Endpoint(
                settings={
                    'response_type': (ProductTableVersion,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/product_table_versions/{product_table_version_id}',
                    'operation_id': 'replace_product_table_version',
                    'http_method': 'PUT',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'product_table_version_id',
                        'product_table_version',
                    ],
                    'required': [
                        'product_table_version_id',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'product_table_version_id':
                            (str,),
                        'product_table_version':
                            (ProductTableVersion,),
                    },
                    'attribute_map': {
                        'product_table_version_id': 'product_table_version_id',
                    },
                    'location_map': {
                        'product_table_version_id': 'path',
                        'product_table_version': 'body',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [
                        'application/json'
                    ]
                },
                api_client=api_client,
                callable=__replace_product_table_version
            )

    create_hypothetical_license_details_query = None 
    create_license = None 
    create_product_table_version = None 
    delete_license = None 
    delete_product_table_version = None 
    get_license = None 
    get_product_table_version = None 
    list_license_details = None 
    list_license_evaluation_contexts = None 
    list_licenses = None 
    list_product_table_versions = None 
    replace_license = None 
    replace_product_table_version = None 



from unittest.mock import MagicMock

class LookupsApiMock:

    def __init__(self):
        self.mock_bulk_query_org_guids = MagicMock()
        self.mock_lookup_org_guid = MagicMock()

    def bulk_query_org_guids(self, *args, **kwargs):
        """
        This method mocks the original api LookupsApi.bulk_query_org_guids with MagicMock.
        """
        return self.mock_bulk_query_org_guids(self, *args, **kwargs)

    def lookup_org_guid(self, *args, **kwargs):
        """
        This method mocks the original api LookupsApi.lookup_org_guid with MagicMock.
        """
        return self.mock_lookup_org_guid(self, *args, **kwargs)




from unittest.mock import MagicMock

class LicensingApiMock:

    def __init__(self):
        self.mock_create_hypothetical_license_details_query = MagicMock()
        self.mock_create_license = MagicMock()
        self.mock_create_product_table_version = MagicMock()
        self.mock_delete_license = MagicMock()
        self.mock_delete_product_table_version = MagicMock()
        self.mock_get_license = MagicMock()
        self.mock_get_product_table_version = MagicMock()
        self.mock_list_license_details = MagicMock()
        self.mock_list_license_evaluation_contexts = MagicMock()
        self.mock_list_licenses = MagicMock()
        self.mock_list_product_table_versions = MagicMock()
        self.mock_replace_license = MagicMock()
        self.mock_replace_product_table_version = MagicMock()

    def create_hypothetical_license_details_query(self, *args, **kwargs):
        """
        This method mocks the original api LicensingApi.create_hypothetical_license_details_query with MagicMock.
        """
        return self.mock_create_hypothetical_license_details_query(self, *args, **kwargs)

    def create_license(self, *args, **kwargs):
        """
        This method mocks the original api LicensingApi.create_license with MagicMock.
        """
        return self.mock_create_license(self, *args, **kwargs)

    def create_product_table_version(self, *args, **kwargs):
        """
        This method mocks the original api LicensingApi.create_product_table_version with MagicMock.
        """
        return self.mock_create_product_table_version(self, *args, **kwargs)

    def delete_license(self, *args, **kwargs):
        """
        This method mocks the original api LicensingApi.delete_license with MagicMock.
        """
        return self.mock_delete_license(self, *args, **kwargs)

    def delete_product_table_version(self, *args, **kwargs):
        """
        This method mocks the original api LicensingApi.delete_product_table_version with MagicMock.
        """
        return self.mock_delete_product_table_version(self, *args, **kwargs)

    def get_license(self, *args, **kwargs):
        """
        This method mocks the original api LicensingApi.get_license with MagicMock.
        """
        return self.mock_get_license(self, *args, **kwargs)

    def get_product_table_version(self, *args, **kwargs):
        """
        This method mocks the original api LicensingApi.get_product_table_version with MagicMock.
        """
        return self.mock_get_product_table_version(self, *args, **kwargs)

    def list_license_details(self, *args, **kwargs):
        """
        This method mocks the original api LicensingApi.list_license_details with MagicMock.
        """
        return self.mock_list_license_details(self, *args, **kwargs)

    def list_license_evaluation_contexts(self, *args, **kwargs):
        """
        This method mocks the original api LicensingApi.list_license_evaluation_contexts with MagicMock.
        """
        return self.mock_list_license_evaluation_contexts(self, *args, **kwargs)

    def list_licenses(self, *args, **kwargs):
        """
        This method mocks the original api LicensingApi.list_licenses with MagicMock.
        """
        return self.mock_list_licenses(self, *args, **kwargs)

    def list_product_table_versions(self, *args, **kwargs):
        """
        This method mocks the original api LicensingApi.list_product_table_versions with MagicMock.
        """
        return self.mock_list_product_table_versions(self, *args, **kwargs)

    def replace_license(self, *args, **kwargs):
        """
        This method mocks the original api LicensingApi.replace_license with MagicMock.
        """
        return self.mock_replace_license(self, *args, **kwargs)

    def replace_product_table_version(self, *args, **kwargs):
        """
        This method mocks the original api LicensingApi.replace_product_table_version with MagicMock.
        """
        return self.mock_replace_product_table_version(self, *args, **kwargs)




from dataclasses import dataclass
from . import (
    apps,
    orgs,
    context,
    issuers as _issuers,
    users as _users,
    resources as _resources,
    connectors as _connectors,
    output,
)


@dataclass
class PrimaryKeyMapping:
    primary_key_field: str
    collection: str = None


def get_key_from_dict(obj, pk_mapping):
    if pk_mapping.collection:
        return obj[pk_mapping.collection][pk_mapping.primary_key_field]
    return obj[pk_mapping.primary_key_field]


def get_key_from_obj(obj, pk_mapping):
    if pk_mapping.collection:
        base = getattr(obj, pk_mapping.collection)
        return getattr(base, pk_mapping.primary_key_field)
    return getattr(obj, pk_mapping.primary_key_field)


def _get_primary_keys(obj, primary_key_mapping):
    res = []
    for mapping in primary_key_mapping:
        if isinstance(obj, dict):
            res.append(get_key_from_dict(obj, mapping))
        else:
            res.append(get_key_from_obj(obj, mapping))
    if len(res) == 1:
        return res[0]
    return tuple(res)


def get_orphaned_resources(working_resource_set, obj_list, primary_key_mapping):
    result = []
    for obj in obj_list:
        key = _get_primary_keys(obj, primary_key_mapping)
        if key not in working_resource_set:
            result.append(obj)
    return result


def get_org_id_list(ctx, enabled=None, **kwargs):
    return {org["id"] for org in orgs.query(ctx, enabled=enabled, **kwargs)}


def get_org_id_map(ctx, enabled=None, **kwargs):
    orgs_map = {}
    for org in orgs.query(ctx, enabled=enabled, **kwargs):
        orgs_map[org["id"]] = org
    return orgs_map


# This function takes a working resource set, query function, and a mapping object
# that indicates how to get the primary key values out of the objects returned from
# the query function. These keys are compared to the working resource set. Keys not
# found in the working resource set are determined to be orphaned.
def get_api_orphaned_resources(
    ctx, working_resource_set, query_func, primary_key_mapping, **kwargs
):
    obj_list = query_func(ctx, **kwargs)
    return (
        get_orphaned_resources(working_resource_set, obj_list, primary_key_mapping),
        obj_list,
    )


def output_orphan_info(ctx, collection_name, orphan_tuple, format_func):
    orphans = orphan_tuple[0]
    num_orphans = len(orphans)
    total = len(orphan_tuple[1])

    if context.output_console(ctx):
        print(f"There are {total} {collection_name}")
        print(f"There are {num_orphans} orphanded {collection_name}")
    print(format_func(ctx, orphans))


RESOURCE_TO_FORMATTER_MAP = {
    "applications": apps.format_apps_for_garbage_collection,
    "issuers": _issuers.format_issuers_for_garbage_collection,
    "resources": _resources.format_permissions,
    "users": _users.format_users_for_garbage_collection,
    "roles": _users.format_org_user_roles,
    "connectors": _connectors.format_for_gc,
}


def output_orphaned_resources(ctx, result_table):
    for k, v in result_table.items():
        output_orphan_info(ctx, k, v, RESOURCE_TO_FORMATTER_MAP[k])


def get_all_orphaned_resources(
    ctx,
    applications=None,
    issuers=None,
    users=None,
    resources=None,
    connectors=None,
    roles=None,
    **kwargs,
):

    result = {}
    if any([applications, issuers, users, resources, connectors]):
        result.update(
            get_org_orphaned_resources(
                ctx,
                applications=applications,
                issuers=issuers,
                users=users,
                resources=resources,
                connectors=connectors,
                **kwargs,
            )
        )

    if roles:
        result.update(get_app_orphaned_resources(ctx, roles=roles, **kwargs))
    return result


def get_org_orphaned_resources(
    ctx,
    applications=None,
    issuers=None,
    users=None,
    resources=None,
    connectors=None,
    mark_connectors_deleted=None,
    only_org_enabled=None,
    dry_run=False,
    **kwargs,
):
    enabled = None
    if only_org_enabled:
        enabled = True

    # retrieve a map of all orgs, so we know the name of things
    # that may be deleted
    org_id_map = get_org_id_map(ctx, **kwargs)
    org_id_list = get_org_id_list(ctx, enabled=enabled, **kwargs)

    result = {}

    output.output_if_console(ctx, f"Total Organisations: {len(org_id_list)}")

    if applications:
        result["applications"] = get_api_orphaned_resources(
            ctx, org_id_list, apps.query, [PrimaryKeyMapping("org_id")], **kwargs
        )

    if issuers:
        result["issuers"] = get_api_orphaned_resources(
            ctx, org_id_list, _issuers.query, [PrimaryKeyMapping("org_id")], **kwargs
        )

    if connectors:
        result["connectors"] = get_api_orphaned_resources(
            ctx,
            org_id_list,
            _connectors.query,
            [PrimaryKeyMapping("org_id", collection="spec")],
            **kwargs,
        )
        if mark_connectors_deleted:
            for connector in result["connectors"][0]:
                org = org_id_map.get(connector["spec"]["org_id"])
                if not org:
                    org_name = "deleted"
                else:
                    org_name = org.get("organisation")
                if dry_run:
                    output.output_if_console(
                        ctx,
                        "would mark connector deleted: "
                        f"{connector['metadata']['id']}, org {org_name}",
                    )
                else:
                    output.output_if_console(
                        ctx,
                        "marking connector deleted: "
                        f"{connector['metadata']['id']}, org {org_name}",
                    )
                    _connectors.replace_agent(
                        ctx,
                        connector["metadata"]["id"],
                        org_id=connector["spec"]["org_id"],
                        admin_status="deleted",
                    )

    if resources:
        result["resources"] = get_api_orphaned_resources(
            ctx,
            org_id_list,
            _resources.query_permissions,
            [PrimaryKeyMapping("org_id", collection="spec")],
            **kwargs,
        )

    if users:
        total = _users.query(ctx, **kwargs)["users"]
        orgless = _users.query(ctx, orgless_users=True, **kwargs)["users"]
        result["users"] = orgless, total

    return result


def get_app_orphaned_resources(ctx, roles=None, **kwargs):
    result = {}
    list_of_apps = apps.query(ctx, **kwargs)
    app_org_id_list = {(app.name, app.org_id) for app in list_of_apps}
    orphans = get_api_orphaned_resources(
        ctx,
        app_org_id_list,
        _users.list_org_user_roles,
        [PrimaryKeyMapping("application"), PrimaryKeyMapping("org_id")],
        **kwargs,
    )

    orphans = [
        orphan
        for orphan in orphans
        if not orphan.application.startswith("urn:api:agilicus")
    ]

    result["roles"] = orphans
    return result
